@@ -37,6 +37,8 @@ static GHashTable *session_hash;
 static connman_bool_t sessionmode;
 static struct connman_session *ecall_session;
 static GSList *policy_list;
+static uint32_t session_mark = 256;
+static struct firewall_context *global_firewall;
 
 enum connman_session_trigger {
 	CONNMAN_SESSION_TRIGGER_UNKNOWN		= 0,
@@ -60,6 +62,7 @@ enum connman_session_state {
 };
 
 struct service_entry {
+	struct connman_session *session;
 	/* track why this service was selected */
 	enum connman_session_reason reason;
 	enum connman_service_state state;
@@ -93,7 +96,14 @@ struct connman_session {
 
 	connman_bool_t ecall;
 
-	GSequence *service_list;
+	enum connman_session_id_type id_type;
+	struct firewall_context *fw;
+	struct nfacct_context *nfctx;
+	uint32_t mark;
+	int index;
+	char *gateway;
+
+	GList *service_list;
 	GHashTable *service_hash;
 };
 
@@ -223,6 +233,7 @@ static char *service2bearer(enum connman_service_type type)
 	case CONNMAN_SERVICE_TYPE_VPN:
 		return "vpn";
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+		return "*";
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
 	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_GADGET:
@@ -232,6 +243,201 @@ static char *service2bearer(enum connman_service_type type)
 	return "";
 }
 
+static int init_firewall(void)
+{
+	struct firewall_context *fw;
+	int err;
+
+	fw = __connman_firewall_create();
+
+	err = __connman_firewall_add_rule(fw, "mangle", "INPUT",
+					"-j CONNMARK --restore-mark");
+	if (err < 0)
+		goto err;
+
+	err = __connman_firewall_add_rule(fw, "mangle", "POSTROUTING",
+					"-j CONNMARK --save-mark");
+	if (err < 0)
+		goto err;
+
+	err = __connman_firewall_enable(fw);
+	if (err < 0)
+		goto err;
+
+	global_firewall = fw;
+
+	return 0;
+
+err:
+	__connman_firewall_destroy(fw);
+
+	return err;
+}
+
+static void cleanup_firewall(void)
+{
+	if (global_firewall == NULL)
+		return;
+
+	__connman_firewall_disable(global_firewall);
+	__connman_firewall_destroy(global_firewall);
+}
+
+static int init_firewall_session(struct connman_session *session)
+{
+	struct firewall_context *fw;
+	int err;
+
+	if (session->policy_config->id_type == CONNMAN_SESSION_ID_TYPE_UNKNOWN)
+		return 0;
+
+	DBG("");
+
+	fw = __connman_firewall_create();
+	if (fw == NULL)
+		return -ENOMEM;
+
+	switch (session->policy_config->id_type) {
+	case CONNMAN_SESSION_ID_TYPE_UID:
+		err = __connman_firewall_add_rule(fw, "mangle", "OUTPUT",
+				"-m owner --uid-owner %s -j MARK --set-mark %d",
+						session->policy_config->id,
+						session->mark);
+		break;
+	case CONNMAN_SESSION_ID_TYPE_GID:
+		err = __connman_firewall_add_rule(fw, "mangle", "OUTPUT",
+				"-m owner --gid-owner %s -j MARK --set-mark %d",
+						session->policy_config->id,
+						session->mark);
+		break;
+	case CONNMAN_SESSION_ID_TYPE_LSM:
+	default:
+		err = -EINVAL;
+	}
+
+	if (err < 0)
+		goto err;
+
+	session->id_type = session->policy_config->id_type;
+
+	err = __connman_firewall_add_rule(fw, "filter", "INPUT",
+		"-m mark --mark %d -m nfacct --nfacct-name session-input-%d",
+		session->mark,
+		session->mark);
+	if (err < 0)
+		goto err;
+
+	err = __connman_firewall_add_rule(fw, "filter", "OUTPUT",
+		"-m mark --mark %d -m nfacct --nfacct-name session-output-%d",
+		session->mark,
+		session->mark);
+	if (err < 0)
+		goto err;
+
+	err = __connman_firewall_enable(fw);
+	if (err)
+		goto err;
+
+	session->fw = fw;
+
+	return 0;
+
+err:
+	__connman_firewall_destroy(fw);
+
+	return err;
+}
+
+static void cleanup_firewall_session(struct connman_session *session)
+{
+	if (session->fw == NULL)
+		return;
+
+	__connman_firewall_disable(session->fw);
+	__connman_firewall_destroy(session->fw);
+
+	session->fw = NULL;
+}
+
+static int init_routing_table(struct connman_session *session)
+{
+	int err;
+
+	if (session->policy_config->id_type == CONNMAN_SESSION_ID_TYPE_UNKNOWN)
+		return 0;
+
+	DBG("");
+
+	err = __connman_inet_add_fwmark_rule(session->mark,
+						AF_INET, session->mark);
+	if (err < 0)
+		return err;
+
+	err = __connman_inet_add_fwmark_rule(session->mark,
+						AF_INET6, session->mark);
+	if (err < 0)
+		__connman_inet_del_fwmark_rule(session->mark,
+						AF_INET, session->mark);
+
+	return err;
+}
+
+static void del_default_route(struct connman_session *session)
+{
+	if (session->gateway != NULL)
+		return;
+
+	DBG("index %d routing table %d default gateway %s",
+		session->index, session->mark, session->gateway);
+
+	__connman_inet_del_default_from_table(session->mark,
+					session->index, session->gateway);
+	g_free(session->gateway);
+	session->gateway = NULL;
+	session->index = -1;
+}
+
+static void add_default_route(struct connman_session *session)
+{
+	struct session_info *info = session->info;
+	struct connman_ipconfig *ipconfig;
+	int err;
+
+	if (info->entry == NULL)
+		return;
+
+	ipconfig = __connman_service_get_ip4config(info->entry->service);
+	session->index = __connman_ipconfig_get_index(ipconfig);
+	session->gateway = g_strdup(__connman_ipconfig_get_gateway(ipconfig));
+
+	DBG("index %d routing table %d default gateway %s",
+		session->index, session->mark, session->gateway);
+
+	err = __connman_inet_add_default_to_table(session->mark,
+					session->index, session->gateway);
+	if (err < 0)
+		DBG("session %p %s", session, strerror(-err));
+}
+
+static void cleanup_routing_table(struct connman_session *session)
+{
+	DBG("");
+
+	__connman_inet_del_fwmark_rule(session->mark,
+					AF_INET6, session->mark);
+
+	__connman_inet_del_fwmark_rule(session->mark,
+					AF_INET, session->mark);
+
+	del_default_route(session);
+}
+
+static void update_routing_table(struct connman_session *session)
+{
+	del_default_route(session);
+	add_default_route(session);
+}
+
 static void destroy_policy_config(struct connman_session *session)
 {
 	if (session->policy == NULL) {
@@ -257,20 +463,21 @@ static void free_session(struct connman_session *session)
 	g_free(session->notify_path);
 	g_free(session->info);
 	g_free(session->info_last);
+	g_free(session->gateway);
 
 	g_free(session);
 }
 
-static void cleanup_session(gpointer user_data)
+static void cleanup_session_final(struct connman_session *session)
 {
-	struct connman_session *session = user_data;
 	struct session_info *info = session->info;
 
 	DBG("remove %s", session->session_path);
 
 	g_slist_free(session->user_allowed_bearers);
-	g_hash_table_destroy(session->service_hash);
-	g_sequence_free(session->service_list);
+	if (session->service_hash != NULL)
+		g_hash_table_destroy(session->service_hash);
+	g_list_free(session->service_list);
 
 	if (info->entry != NULL &&
 			info->entry->reason == CONNMAN_SESSION_REASON_CONNECT) {
@@ -280,6 +487,34 @@ static void cleanup_session(gpointer user_data)
 	free_session(session);
 }
 
+static void nfacct_cleanup_cb(unsigned int error, struct nfacct_context *ctx,
+				void *user_data)
+{
+	struct connman_session *session = user_data;
+
+	DBG("");
+
+	__connman_nfacct_destroy_context(session->nfctx);
+
+	cleanup_session_final(session);
+}
+
+static void cleanup_session(gpointer user_data)
+{
+	struct connman_session *session = user_data;
+
+	DBG("remove %s", session->session_path);
+
+	cleanup_routing_table(session);
+	cleanup_firewall_session(session);
+
+	if (session->nfctx != NULL)
+		__connman_nfacct_disable(session->nfctx, nfacct_cleanup_cb,
+						session);
+	else
+		cleanup_session_final(session);
+}
+
 static int assign_policy_plugin(struct connman_session *session)
 {
 	if (session->policy != NULL)
@@ -293,28 +528,30 @@ static int assign_policy_plugin(struct connman_session *session)
 	return 0;
 }
 
-struct user_config {
+struct creation_data {
 	DBusMessage *pending;
+	struct connman_session *session;
 
+	/* user config */
 	enum connman_session_type type;
 	GSList *allowed_bearers;
 };
 
-static void cleanup_user_config(struct user_config *user_config)
+static void cleanup_creation_data(struct creation_data *creation_data)
 {
-	if (user_config == NULL)
+	if (creation_data == NULL)
 		return;
 
-	if (user_config->pending != NULL)
-		dbus_message_unref(user_config->pending);
+	if (creation_data->pending != NULL)
+		dbus_message_unref(creation_data->pending);
 
-	g_slist_free(user_config->allowed_bearers);
-	g_free(user_config);
+	g_slist_free(creation_data->allowed_bearers);
+	g_free(creation_data);
 }
 
 static int create_policy_config(struct connman_session *session,
 				connman_session_config_func_t cb,
-				struct user_config *user_config)
+				struct creation_data *creation_data)
 {
 	struct connman_session_config *config;
 
@@ -322,14 +559,14 @@ static int create_policy_config(struct connman_session *session,
 		config = connman_session_create_default_config();
 		if (config == NULL) {
 			free_session(session);
-			cleanup_user_config(user_config);
+			cleanup_creation_data(creation_data);
 			return -ENOMEM;
 		}
 
-		return cb(session, config, user_config, 0);
+		return cb(session, config, creation_data, 0);
 	}
 
-	return (*session->policy->create)(session, cb, user_config);
+	return (*session->policy->create)(session, cb, creation_data);
 }
 
 static void probe_policy(struct connman_session_policy *policy)
@@ -412,6 +649,10 @@ void connman_session_policy_unregister(struct connman_session_policy *policy)
 
 void connman_session_set_default_config(struct connman_session_config *config)
 {
+	config->id_type = CONNMAN_SESSION_ID_TYPE_UNKNOWN;
+	g_free(config->id);
+	config->id = NULL;
+
 	config->priority = FALSE;
 	config->roaming_policy = CONNMAN_SESSION_ROAMING_POLICY_DEFAULT;
 	config->type = CONNMAN_SESSION_TYPE_ANY;
@@ -1123,6 +1364,8 @@ static void deselect_and_disconnect(struct connman_session *session)
 	deselect_service(info);
 
 	info->reason = CONNMAN_SESSION_REASON_FREE_RIDE;
+
+	update_routing_table(session);
 }
 
 static void select_connected_service(struct session_info *info,
@@ -1169,6 +1412,8 @@ static void select_service(struct session_info *info,
 		select_connected_service(info, entry);
 	else
 		select_offline_service(info, entry);
+
+	update_routing_table(entry->session);
 }
 
 static void select_and_connect(struct connman_session *session,
@@ -1176,16 +1421,14 @@ static void select_and_connect(struct connman_session *session,
 {
 	struct session_info *info = session->info;
 	struct service_entry *entry = NULL;
-	GSequenceIter *iter;
+	GList *list;
 
 	DBG("session %p reason %s", session, reason2string(reason));
 
 	info->reason = reason;
 
-	iter = g_sequence_get_begin_iter(session->service_list);
-
-	while (g_sequence_iter_is_end(iter) == FALSE) {
-		entry = g_sequence_get(iter);
+	for (list = session->service_list; list != NULL; list = list->next) {
+		entry = list->data;
 
 		switch (entry->state) {
 		case CONNMAN_SERVICE_STATE_ASSOCIATION:
@@ -1200,12 +1443,11 @@ static void select_and_connect(struct connman_session *session,
 		case CONNMAN_SERVICE_STATE_FAILURE:
 			break;
 		}
-
-		iter = g_sequence_iter_next(iter);
 	}
 }
 
-static struct service_entry *create_service_entry(struct connman_service *service,
+static struct service_entry *create_service_entry(struct connman_session * session,
+					struct connman_service *service,
 					const char *name,
 					enum connman_service_state state)
 {
@@ -1233,12 +1475,39 @@ static struct service_entry *create_service_entry(struct connman_service *servic
 	type = connman_service_get_type(entry->service);
 	entry->bearer = service2bearer(type);
 
+	entry->session = session;
+
 	return entry;
 }
 
+static void iterate_service_cb(struct connman_service *service,
+				const char *name,
+				enum connman_service_state state,
+				void *user_data)
+{
+	struct connman_session *session = user_data;
+	struct service_entry *entry;
+
+	if (service_match(session, service) == FALSE)
+		return;
+
+	entry = create_service_entry(session, service, name, state);
+	if (entry == NULL)
+		return;
+
+	g_hash_table_replace(session->service_hash, service, entry);
+}
+
 static void destroy_service_entry(gpointer data)
 {
 	struct service_entry *entry = data;
+	struct session_info *info = entry->session->info;
+	struct connman_session *session;
+
+	if (info != NULL && info->entry == entry) {
+		session = entry->session;
+		deselect_and_disconnect(session);
+	}
 
 	pending_timeout_remove_all(entry);
 	g_free(entry->ifname);
@@ -1248,33 +1517,28 @@ static void destroy_service_entry(gpointer data)
 
 static void populate_service_list(struct connman_session *session)
 {
-	struct service_entry *entry;
-	GSequenceIter *iter;
+	GHashTableIter iter;
+	gpointer key, value;
 
 	session->service_hash =
 		g_hash_table_new_full(g_direct_hash, g_direct_equal,
-					NULL, NULL);
-	session->service_list = __connman_service_get_list(session,
-							service_match,
-							create_service_entry,
-							destroy_service_entry);
+					NULL, destroy_service_entry);
+	__connman_service_iterate_services(iterate_service_cb, session);
 
-	g_sequence_sort(session->service_list, sort_services, session);
+	g_hash_table_iter_init(&iter, session->service_hash);
 
-	iter = g_sequence_get_begin_iter(session->service_list);
-
-	while (g_sequence_iter_is_end(iter) == FALSE) {
-		entry = g_sequence_get(iter);
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		struct service_entry *entry = value;
 
 		DBG("service %p type %s name %s", entry->service,
 			service2bearer(connman_service_get_type(entry->service)),
 			entry->name);
-
-		g_hash_table_replace(session->service_hash,
-					entry->service, iter);
-
-		iter = g_sequence_iter_next(iter);
+		session->service_list = g_list_prepend(session->service_list,
+							entry);
 	}
+
+	session->service_list = g_list_sort_with_data(session->service_list,
+						sort_services, session);
 }
 
 static void session_changed(struct connman_session *session,
@@ -1282,8 +1546,8 @@ static void session_changed(struct connman_session *session,
 {
 	struct session_info *info = session->info;
 	struct session_info *info_last = session->info_last;
-	GSequenceIter *service_iter = NULL, *service_iter_last = NULL;
-	GSequence *service_list_last;
+	struct connman_service *service;
+	struct service_entry *entry = NULL, *entry_last = NULL;
 	GHashTable *service_hash_last;
 
 	/*
@@ -1311,20 +1575,26 @@ static void session_changed(struct connman_session *session,
 		if (info->config.allowed_bearers != info_last->config.allowed_bearers) {
 
 			service_hash_last = session->service_hash;
-			service_list_last = session->service_list;
+			g_list_free(session->service_list);
+			session->service_list = NULL;
+
+			if (info->entry != NULL)
+				service = info->entry->service;
+			else
+				service = NULL;
 
 			populate_service_list(session);
 
 			if (info->entry != NULL) {
-				service_iter_last = g_hash_table_lookup(
+				entry_last = g_hash_table_lookup(
 							service_hash_last,
 							info->entry->service);
-				service_iter = g_hash_table_lookup(
+				entry = g_hash_table_lookup(
 							session->service_hash,
-							info->entry->service);
+							service);
 			}
 
-			if (service_iter == NULL && service_iter_last != NULL) {
+			if (entry == NULL && entry_last != NULL) {
 				/*
 				 * The currently selected service is
 				 * not part of this session anymore.
@@ -1332,8 +1602,7 @@ static void session_changed(struct connman_session *session,
 				deselect_and_disconnect(session);
 			}
 
-			g_hash_table_remove_all(service_hash_last);
-			g_sequence_free(service_list_last);
+			g_hash_table_destroy(service_hash_last);
 		}
 
 		if (info->config.type != info_last->config.type) {
@@ -1408,6 +1677,17 @@ int connman_session_config_update(struct connman_session *session)
 	 * might have changed. We can still optimize this later.
 	 */
 
+	if (session->id_type != session->policy_config->id_type) {
+		cleanup_firewall_session(session);
+		err = init_firewall_session(session);
+		if (err < 0) {
+			connman_session_destroy(session);
+			return err;
+		}
+
+		session->id_type = session->policy_config->id_type;
+	}
+
 	err = apply_policy_on_bearers(
 		session->policy_config->allowed_bearers,
 		session->user_allowed_bearers,
@@ -1619,20 +1899,25 @@ static const GDBusMethodTable session_methods[] = {
 	{ },
 };
 
-static int session_create_cb(struct connman_session *session,
-				struct connman_session_config *config,
-				void *user_data, int err)
+static void session_nfacct_stats_cb(struct nfacct_context *ctx,
+					uint64_t packets, uint64_t bytes,
+					void *user_data)
+{
+	struct connman_session *session = user_data;
+
+	DBG("session %p", session);
+
+	/* XXX add accounting code here */
+}
+
+static int session_create_final(struct creation_data *creation_data,
+				struct connman_session *session)
 {
-	DBusMessage *reply;
-	struct user_config *user_config = user_data;
 	struct session_info *info, *info_last;
+	DBusMessage *reply;
+	int err;
 
-	DBG("session %p config %p", session, config);
-
-	if (err != 0)
-		goto out;
-
-	session->policy_config = config;
+	DBG("");
 
 	info = session->info;
 	info_last = session->info_last;
@@ -1643,20 +1928,20 @@ static int session_create_cb(struct connman_session *session,
 	info->state = CONNMAN_SESSION_STATE_DISCONNECTED;
 	info->config.type = apply_policy_on_type(
 				session->policy_config->type,
-				user_config->type);
+				creation_data->type);
 	info->config.priority = session->policy_config->priority;
 	info->config.roaming_policy = session->policy_config->roaming_policy;
 	info->entry = NULL;
 
-	session->user_allowed_bearers = user_config->allowed_bearers;
-	user_config->allowed_bearers = NULL;
+	session->user_allowed_bearers = creation_data->allowed_bearers;
+	creation_data->allowed_bearers = NULL;
 
 	err = apply_policy_on_bearers(
 			session->policy_config->allowed_bearers,
 			session->user_allowed_bearers,
 			&info->config.allowed_bearers);
 	if (err < 0)
-		goto out;
+		goto err;
 
 	g_hash_table_replace(session_hash, session->session_path, session);
 
@@ -1669,14 +1954,14 @@ static int session_create_cb(struct connman_session *session,
 		connman_error("Failed to register %s", session->session_path);
 		g_hash_table_remove(session_hash, session->session_path);
 		err = -EINVAL;
-		goto out;
+		goto err;
 	}
 
-	reply = g_dbus_create_reply(user_config->pending,
+	reply = g_dbus_create_reply(creation_data->pending,
 				DBUS_TYPE_OBJECT_PATH, &session->session_path,
 				DBUS_TYPE_INVALID);
 	g_dbus_send_message(connection, reply);
-	user_config->pending = NULL;
+	creation_data->pending = NULL;
 
 	populate_service_list(session);
 
@@ -1690,15 +1975,106 @@ static int session_create_cb(struct connman_session *session,
 
 	session_changed(session, CONNMAN_SESSION_TRIGGER_SETTING);
 
-out:
-	if (err < 0) {
-		reply = __connman_error_failed(user_config->pending, -err);
-		g_dbus_send_message(connection, reply);
+	cleanup_creation_data(creation_data);
 
-		free_session(session);
+err:
+	return err;
+}
+
+static void session_nfacct_enable_cb(unsigned int error,
+					struct nfacct_context *ctx,
+					void *user_data)
+{
+	struct creation_data *creation_data = user_data;
+	struct connman_session *session = creation_data->session;
+	DBusMessage *reply;
+	int err;
+
+	DBG("");
+
+	if (error != 0) {
+		err = -error;
+		goto err;
 	}
 
-	cleanup_user_config(user_config);
+	err = init_firewall_session(session);
+	if (err < 0)
+		goto err;
+
+	err = init_routing_table(session);
+	if (err < 0)
+		goto err;
+
+	err = session_create_final(creation_data, session);
+	if (err < 0)
+		goto err;
+
+	return;
+
+err:
+	reply = __connman_error_failed(creation_data->pending, -err);
+	g_dbus_send_message(connection, reply);
+	creation_data->pending = NULL;
+
+	cleanup_session(creation_data->session);
+	cleanup_creation_data(creation_data);
+}
+
+static int session_policy_config_cb(struct connman_session *session,
+				struct connman_session_config *config,
+				void *user_data, int err)
+{
+	struct creation_data *creation_data = user_data;
+	char *input, *output;
+	DBusMessage *reply;
+
+	DBG("session %p config %p", session, config);
+
+	if (err < 0)
+		goto err;
+
+	session->policy_config = config;
+
+	session->mark = session_mark++;
+	session->index = -1;
+
+	session->nfctx = __connman_nfacct_create_context();
+	if (session->nfctx == NULL) {
+		err = -ENOMEM;
+		goto err;
+	}
+
+	input = g_strdup_printf("session-input-%d", session->mark);
+	err = __connman_nfacct_add(session->nfctx, input,
+					session_nfacct_stats_cb,
+					session);
+	g_free(input);
+	if (err < 0)
+		goto err;
+
+	output = g_strdup_printf("session-output-%d", session->mark);
+	err = __connman_nfacct_add(session->nfctx, output,
+					session_nfacct_stats_cb,
+					session);
+	g_free(output);
+	if (err < 0)
+		goto err;
+
+	err = __connman_nfacct_enable(session->nfctx,
+					session_nfacct_enable_cb,
+					creation_data);
+	if (err < 0)
+		goto err;
+
+	return -EINPROGRESS;
+
+err:
+	reply = __connman_error_failed(creation_data->pending, -err);
+	g_dbus_send_message(connection, reply);
+	creation_data->pending = NULL;
+
+	cleanup_session(session);
+	cleanup_creation_data(creation_data);
 
 	return err;
 }
@@ -1709,7 +2085,7 @@ int __connman_session_create(DBusMessage *msg)
 	char *session_path = NULL;
 	DBusMessageIter iter, array;
 	struct connman_session *session = NULL;
-	struct user_config *user_config = NULL;
+	struct creation_data *creation_data = NULL;
 	connman_bool_t user_allowed_bearers = FALSE;
 	connman_bool_t user_connection_type = FALSE;
 	int err;
@@ -1727,13 +2103,13 @@ int __connman_session_create(DBusMessage *msg)
 		goto err;
 	}
 
-	user_config = g_try_new0(struct user_config, 1);
-	if (user_config == NULL) {
+	creation_data = g_try_new0(struct creation_data, 1);
+	if (creation_data == NULL) {
 		err = -ENOMEM;
 		goto err;
 	}
 
-	user_config->pending = dbus_message_ref(msg);
+	creation_data->pending = dbus_message_ref(msg);
 
 	dbus_message_iter_init(msg, &iter);
 	dbus_message_iter_recurse(&iter, &array);
@@ -1752,7 +2128,7 @@ int __connman_session_create(DBusMessage *msg)
 		case DBUS_TYPE_ARRAY:
 			if (g_str_equal(key, "AllowedBearers") == TRUE) {
 				err = parse_bearers(&value,
-						&user_config->allowed_bearers);
+						&creation_data->allowed_bearers);
 				if (err < 0)
 					goto err;
 
@@ -1765,7 +2141,7 @@ int __connman_session_create(DBusMessage *msg)
 		case DBUS_TYPE_STRING:
 			if (g_str_equal(key, "ConnectionType") == TRUE) {
 				dbus_message_iter_get_basic(&value, &val);
-				user_config->type =
+				creation_data->type =
 					connman_session_parse_connection_type(val);
 
 				user_connection_type = TRUE;
@@ -1784,10 +2160,10 @@ int __connman_session_create(DBusMessage *msg)
 	 * For AllowedBearers this is '*', ...
 	 */
 	if (user_allowed_bearers == FALSE) {
-		user_config->allowed_bearers =
+		creation_data->allowed_bearers =
 			g_slist_append(NULL,
 				GINT_TO_POINTER(CONNMAN_SERVICE_TYPE_UNKNOWN));
-		if (user_config->allowed_bearers == NULL) {
+		if (creation_data->allowed_bearers == NULL) {
 			err = -ENOMEM;
 			goto err;
 		}
@@ -1795,7 +2171,7 @@ int __connman_session_create(DBusMessage *msg)
 
 	/* ... and for ConnectionType it is 'any'. */
 	if (user_connection_type == FALSE)
-		user_config->type = CONNMAN_SESSION_TYPE_ANY;
+		creation_data->type = CONNMAN_SESSION_TYPE_ANY;
 
 	dbus_message_iter_next(&iter);
 	dbus_message_iter_get_basic(&iter, &notify_path);
@@ -1826,6 +2202,7 @@ int __connman_session_create(DBusMessage *msg)
 		goto err;
 	}
 
+	creation_data->session = session;
 	session->session_path = session_path;
 
 	session->info = g_try_new0(struct session_info, 1);
@@ -1850,7 +2227,8 @@ int __connman_session_create(DBusMessage *msg)
 	if (err < 0)
 		goto err;
 
-	err = create_policy_config(session, session_create_cb, user_config);
+	err = create_policy_config(session, session_policy_config_cb,
+					creation_data);
 	if (err < 0 && err != -EINPROGRESS)
 		return err;
 
@@ -1861,7 +2239,7 @@ err:
 
 	free_session(session);
 
-	cleanup_user_config(user_config);
+	cleanup_creation_data(creation_data);
 	return err;
 }
 
@@ -1923,7 +2301,6 @@ static void service_add(struct connman_service *service,
 			const char *name)
 {
 	GHashTableIter iter;
-	GSequenceIter *iter_service_list;
 	gpointer key, value;
 	struct connman_session *session;
 	struct service_entry *entry;
@@ -1938,18 +2315,15 @@ static void service_add(struct connman_service *service,
 		if (service_match(session, service) == FALSE)
 			continue;
 
-		entry = create_service_entry(service, name,
+		entry = create_service_entry(session, service, name,
 						CONNMAN_SERVICE_STATE_IDLE);
 		if (entry == NULL)
 			continue;
 
-		iter_service_list =
-			g_sequence_insert_sorted(session->service_list,
-							entry, sort_services,
-							session);
+		session->service_list = g_list_insert_sorted_with_data(
+			session->service_list, entry, sort_services, session);
 
-		g_hash_table_replace(session->service_hash, service,
-					iter_service_list);
+		g_hash_table_replace(session->service_hash, service, entry);
 
 		session_changed(session, CONNMAN_SESSION_TRIGGER_SERVICE);
 	}
@@ -1968,15 +2342,18 @@ static void service_remove(struct connman_service *service)
 	g_hash_table_iter_init(&iter, session_hash);
 
 	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
-		GSequenceIter *seq_iter;
+		struct service_entry *entry;
 		session = value;
 		info = session->info;
 
-		seq_iter = g_hash_table_lookup(session->service_hash, service);
-		if (seq_iter == NULL)
+		entry = g_hash_table_lookup(session->service_hash, service);
+		if (entry == NULL)
 			continue;
 
-		g_sequence_remove(seq_iter);
+		session->service_list = g_list_remove(session->service_list,
+							entry);
+
+		g_hash_table_remove(session->service_hash, service);
 
 		if (info->entry != NULL && info->entry->service == service)
 			info->entry = NULL;
@@ -1996,15 +2373,11 @@ static void service_state_changed(struct connman_service *service,
 
 	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
 		struct connman_session *session = value;
-		GSequenceIter *service_iter;
+		struct service_entry *entry;
 
-		service_iter = g_hash_table_lookup(session->service_hash, service);
-		if (service_iter != NULL) {
-			struct service_entry *entry;
-
-			entry = g_sequence_get(service_iter);
+		entry = g_hash_table_lookup(session->service_hash, service);
+		if (entry != NULL)
 			entry->state = state;
-		}
 
 		session_changed(session,
 				CONNMAN_SESSION_TRIGGER_SERVICE);
@@ -2050,12 +2423,26 @@ static struct connman_notifier session_notifier = {
 	.ipconfig_changed	= ipconfig_changed,
 };
 
+static int session_nfacct_flush_cb(unsigned int error, void *user_data)
+{
+	if (error == 0)
+		return 0;
+
+	connman_error("Could not flush nfacct table: %s", strerror(error));
+
+	return -error;
+}
+
 int __connman_session_init(void)
 {
 	int err;
 
 	DBG("");
 
+	err = init_firewall();
+	if (err < 0)
+		return err;
+
 	connection = connman_dbus_get_connection();
 	if (connection == NULL)
 		return -1;
@@ -2063,6 +2450,7 @@ int __connman_session_init(void)
 	err = connman_notifier_register(&session_notifier);
 	if (err < 0) {
 		dbus_connection_unref(connection);
+		cleanup_firewall();
 		return err;
 	}
 
@@ -2070,6 +2458,9 @@ int __connman_session_init(void)
 						NULL, cleanup_session);
 
 	sessionmode = FALSE;
+
+	__connman_nfacct_flush(session_nfacct_flush_cb, NULL);
+
 	return 0;
 }
 
@@ -2080,6 +2471,8 @@ void __connman_session_cleanup(void)
 	if (connection == NULL)
 		return;
 
+	cleanup_firewall();
+
 	connman_notifier_unregister(&session_notifier);
 
 	g_hash_table_foreach(session_hash, release_session, NULL);
